from datetime import datetime, time, date
from pandas.api.types import CategoricalDtype
from typing import Optional
from .regex_patterns import (
    DATE_REGEX_PATTERNS,
    TIME_REGEX_PATTERNS,
    DATE_TIME_REGEX_PATTERNS,
)
import numpy as np
import pandas as pd
import re


class NormalizeDateUtils:
    # Converter um valor de data string em datetime
    @staticmethod
    def _convert_date_string_to_datetime(date_str: str) -> Optional[datetime]:
        '''
        Converte a data de um dataframe em um único formato datetime
        '''

        for pattern, date_format in DATE_REGEX_PATTERNS:
            if re.match(pattern, date_str):
                try:
                    return datetime.strptime(date_str, date_format)
                except ValueError:
                    continue

        return None
    

    # Converter uma data string para string no formato 'yyyy-mm-dd'
    @staticmethod
    def _convert_str_date_to_date_standard_str(date_str: str) -> Optional[str]:
        '''
        Converte uma data em diferentes formatos para uma string no formato yyyy-mm-dd.
        Retorna None se não conseguir converter.
        '''
        for pattern, date_format in DATE_REGEX_PATTERNS:
            if re.match(pattern, date_str):
                try:
                    dt = datetime.strptime(date_str, date_format)
                    return dt.strftime('%Y-%m-%d')
                except ValueError:
                    continue

        return None


    # Converter todas as linhas de um current_df em datetime
    @staticmethod
    def normalize_str_date_to_date_standard_str_by_columns(
        current_df: pd.DataFrame,
        columns: list[str]
    ) -> None:
        '''
        Converte todas as linhas de uma coluna de data datetime
        '''

        for column in columns:
            if column not in current_df.columns:
                continue
            current_df[column] = (
                current_df[column]
                .apply(lambda row: NormalizeDateUtils._convert_str_date_to_date_standard_str(row))
            )


    # Converter todas as linhas de um current_df em datetime
    @staticmethod
    def normalize_str_date_to_datetime_by_columns(
        current_df: pd.DataFrame,
        columns: list[str]
    ) -> None:
        '''
        Converte todas as linhas de uma coluna de data datetime
        '''

        for column in columns:
            if column not in current_df.columns:
                continue
            current_df[column] = (
                current_df[column]
                .apply(lambda row: NormalizeDateUtils._convert_date_string_to_datetime(row))
            )


    @staticmethod
    def create_column_datetime_from_date(
        current_df: pd.DataFrame,
        date_column: str = 'DATE',
        new_column: str = 'DATETIME',
        before_column: Optional[str] = None
    ) -> None:
        '''
        Cria uma nova coluna `DATETIME` a partir de uma coluna de data
        em formato string "yyyy-mm-dd" ou datetime.date/datetime64.

        Se `before_column` for fornecida e existir no DataFrame,
        a nova coluna será inserida antes dela.
        '''

        def parse_date(row: pd.Series) -> Optional[pd.Timestamp]:
            date_val = row.get(date_column)

            if pd.isna(date_val):
                return np.nan

            try:
                if isinstance(date_val, str):
                    return datetime.strptime(date_val, '%Y-%m-%d')
                elif isinstance(date_val, (datetime, pd.Timestamp)):
                    return pd.to_datetime(date_val)
                elif isinstance(date_val, date):
                    return datetime.combine(date_val, time(0, 0))
                else:
                    return np.nan
            except Exception:
                return np.nan

        result_series: pd.Series = current_df.apply(parse_date, axis=1)

        if before_column and before_column in current_df.columns:
            insert_position: int = current_df.columns.get_loc(before_column)
            current_df.insert(loc=insert_position, column=new_column, value=result_series)
        else:
            current_df[new_column] = result_series



    # TODO: Ver melhor isso
    @staticmethod
    def create_column_datetime_from_date_and_time(
        current_df: pd.DataFrame,
        date_column: str = 'DATE',
        time_column: str = 'TIME',
        new_column: str = 'DATETIME',
        before_column: Optional[str] = None
    ) -> None:
        '''
        Cria uma nova coluna datetime combinando:
        - Uma coluna de data (no formato "yyyy-mm-dd" como string ou datetime)
        - Uma coluna de hora (no formato "HH:MM:SS")

        Se `before_column` for fornecida e existir no DataFrame,
        a nova coluna será inserida antes dela.
        '''

        def combine(row: pd.Series) -> pd.Timestamp:
            date_val = row.get(date_column)
            time_val = row.get(time_column)

            if pd.isna(date_val) or pd.isna(time_val):
                return np.nan

            try:
                if isinstance(date_val, str):
                    date_val = datetime.strptime(date_val, '%Y-%m-%d').date()

                if isinstance(time_val, str):
                    time_parts = [int(part) for part in time_val.strip().split(':')]
                    while len(time_parts) < 3:
                        time_parts.append(0)
                    hour_val = time(*time_parts[:3])
                else:
                    return np.nan

                return datetime.combine(date_val, hour_val)
            except Exception:
                return np.nan

        result_series = current_df.apply(combine, axis=1)

        if before_column and before_column in current_df.columns:
            insert_position = current_df.columns.get_loc(before_column)
            current_df.insert(insert_position, new_column, result_series)
        else:
            current_df[new_column] = result_series



    @staticmethod
    def _normalize_str_time_to_hh_mm_ss(time_str: str) -> Optional[str]:
        '''
        Normaliza uma string de hora para o formato hh:mm:00
        '''
        if not isinstance(time_str, str) or not time_str.strip():
            return np.nan

        time_str = time_str.strip()

        try:
            if re.match(r'^\d{1,2}$', time_str):
                return f'{int(time_str):02d}:00:00'

            if re.match(r'^\d{1,2}:\d{2}$', time_str):
                dt = datetime.strptime(time_str, '%H:%M')
                return dt.strftime('%H:%M:00')

            if re.match(r'^\d{1,2}:\d{2}:\d{2}(\.\d+)?$', time_str):
                dt = datetime.strptime(time_str.split('.')[0], '%H:%M:%S')
                return dt.strftime('%H:%M:%S')
            
            if re.match(r'^\d{4}\sUTC$', time_str):
                dt = datetime.strptime(time_str, '%H%M UTC')
                return dt.strftime('%H:%M:00')
        except ValueError:
            pass

        return np.nan


    @staticmethod
    def normalize_str_time_to_hh_mm_ss_by_columns(
        current_df: pd.DataFrame,
        columns: list[str]
    ) -> None:
        '''
        Aplica a normalização do formato de hora hh:mm:00 em múltiplas colunas do DataFrame.
        Substitui valores inválidos por NaN.
        '''
        for column in columns:
            if column in current_df.columns:
                current_df[column] = current_df[column].apply(
                    NormalizeDateUtils._normalize_str_time_to_hh_mm_ss
                )


    # @staticmethod
    # def create_column_hour_string_by_column(
    #     current_df: pd.DataFrame,
    #     column: Optional[str] = 'DATE',
    #     before_column: Optional[str] = None
    # ) -> None:
    #     '''
    #     Cria uma coluna `HORA` no formato HH:00 a partir de uma coluna datetime.
    #     Se `before_column` for fornecida e existir no DataFrame, a nova coluna será inserida antes dela.
    #     '''

    #     hora_series = current_df[column].dt.hour.apply(lambda h: f'{h:02d}:00')

    #     if before_column and before_column in current_df.columns:
    #         insert_position = current_df.columns.get_loc(before_column)
    #         current_df.insert(insert_position, 'HORA', hora_series)
    #     else:
    #         current_df['HORA'] = hora_series


    def create_column_hora(
        current_df: pd.DataFrame,
        datetime_column: Optional[str] = 'DATETIME',
        before_column: Optional[str] = None
    ) -> None:
        '''
        Cria uma nova coluna HORA no formato "HH:00" (ex: 00:00, 01:00, ..., 23:00)
        a partir de uma coluna datetime. A coluna é transformada em categoria ordenada.
        '''

        horas_ordem: list[str] = [f'{h:02d}:00' for h in range(24)]
        categoria_horas = CategoricalDtype(categories=horas_ordem, ordered=True)

        result_series = current_df[datetime_column].dt.hour.map(lambda h: f'{h:02d}:00')

        if before_column and before_column in current_df.columns:
            insert_position = current_df.columns.get_loc(before_column)
            current_df.insert(insert_position, 'HORA', result_series.astype(categoria_horas))
        else:
            current_df['HORA'] = result_series.astype(categoria_horas)


    def create_column_ano(
        current_df: pd.DataFrame,
        column: Optional[str]
    ) -> None:
        '''
        Cria uma coluna `ANO` com o ano extraído de uma coluna datetime
        '''
        if not column:
            column = 'DATE'

        current_df['ANO'] = current_df[column].dt.year


    @staticmethod
    def create_column_mes_int(current_df: pd.DataFrame, column: Optional[str]) -> None:
        '''
        Cria uma coluna `MES_INT` com o mês extraído de uma coluna datetime.
        '''
        if not column:
            column = 'DATE'

        current_df['MES_INT'] = current_df[column].dt.month

    
    @staticmethod
    def create_column_dia_semana(
        current_df: pd.DataFrame,
        datetime_column: Optional[str] = 'DATETIME',
        before_column: Optional[str] = None
    ) -> None:
        '''
        Cria uma nova coluna DIA_SEMANA com o nome do dia da semana a partir de uma coluna datetime,
        e a transforma em categoria ordenada (SEGUNDA a domingo).
        '''
        ordem_dias: list[str] = [
            'SEGUNDA', 'TERÇA', 'QUARTA', 'QUINTA',
            'SEXTA', 'SABADO', 'DOMINGO'
        ]
        categoria_dias = CategoricalDtype(categories=ordem_dias, ordered=True)

        result_series = current_df[datetime_column].dt.dayofweek.map(lambda i: ordem_dias[i])

        if before_column and before_column in current_df.columns:
            insert_position = current_df.columns.get_loc(before_column)
            current_df.insert(insert_position, 'DIA_SEMANA', result_series.astype(categoria_dias))
        else:
            current_df['DIA_SEMANA'] = result_series.astype(categoria_dias)


    @staticmethod
    def create_column_mes(
        current_df: pd.DataFrame,
        datetime_column: Optional[str] = 'DATETIME',
        before_column: Optional[str] = None
    ) -> None:
        '''
        Cria uma nova coluna MES com o nome do dia da semana a partir de uma coluna datetime,
        e a transforma em categoria ordenada (SEGUNDA a domingo).
        '''
        if not datetime_column:
            datetime_column: str = 'DATE'

        meses_ordem: list[str] = [
            'JAN', 'FEV', 'MAR', 'ABR', 'MAI', 'JUN',
            'JUL', 'AGO', 'SET', 'OUT', 'NOV', 'DEZ'
        ]

        categoria_dias = CategoricalDtype(categories=meses_ordem, ordered=True)

        result_series = current_df[datetime_column].dt.month.map(lambda i: meses_ordem[i - 1])

        if before_column and before_column in current_df.columns:
            insert_position = current_df.columns.get_loc(before_column)
            current_df.insert(insert_position, 'MES', result_series.astype(categoria_dias))
        else:
            current_df['MES'] = result_series.astype(categoria_dias)


    # @staticmethod
    # def create_column_hour_category(
    #     current_df: pd.DataFrame,
    #     time_column: str,
    #     before_column: Optional[str] = None
    # ) -> None:
    #     '''
    #     Cria uma nova coluna "HORA" com a hora extraída (00 a 23) da coluna time_column,
    #     no formato string pyarrow categorizada e ordenada.
    #     '''

    #     # Validação básica
    #     if time_column not in current_df.columns:
    #         raise ValueError(f"Coluna '{time_column}' não encontrada no DataFrame")

    #     # Lista de categorias string zero-padded "00" a "23"
    #     horas_ordem: list[str] = [f"{h:02d}" for h in range(24)]

    #     # Define categoria ordenada
    #     categoria_horas: CategoricalDtype = CategoricalDtype(categories=horas_ordem, ordered=True)

    #     # Extrai substring hh da coluna string[pyarrow] via pyarrow compute
    #     substr_array = pc.utf8_slice(current_df[time_column].array, 0, 2)

    #     # Converte para pd.Series com dtype string[pyarrow]
    #     horas_pyarrow: pd.Series = pd.Series(substr_array).astype("string[pyarrow]")

    #     # Converte para categoria ordenada (pandas Categorical)
    #     horas_categ = horas_pyarrow.astype(categoria_horas)

    #     # Insere a coluna no DataFrame no local correto
    #     if before_column and before_column in current_df.columns:
    #         pos = current_df.columns.get_loc(before_column)
    #         current_df.insert(pos, 'HORA', horas_categ)
    #     else:
    #         current_df['HORA'] = horas_categ


from .regex_patterns import (
    RAW_INVALID_VALUES
)
import numpy as np
import pandas as pd


class NormalizeNullUtils:
    @staticmethod
    def _is_nan_value(value: str, raw_invalid_values: list) -> bool:

        if pd.isna(value) or value in raw_invalid_values:
            return True
        else:
            return False

    @staticmethod
    def normalize_null_by_columns(
        current_df: pd.DataFrame,
        columns: list[str],
        # null_values: list[str] = ['-9999', -9999]
        null_values: list[str] = []
    ) -> None:
        
        raw_invalid_values: list = RAW_INVALID_VALUES.copy()

        if null_values:
            raw_invalid_values.extend(null_values)

        for column in columns:
            if column not in current_df.columns:
                continue

            # Maior overhead, reescreve linha com valor original se não estiver na regra
            current_df[column] = current_df[column].apply(
                lambda row: pd.NA if NormalizeNullUtils._is_nan_value(row, raw_invalid_values) else row
            )

            # Menor overhead, não reescreve linha que não está na regra
            # mask: pd.Series = current_df[column].apply(NormalizeNullUtils._is_nan_value)
            # current_df.loc[mask, column] = np.nan



from .analysis_utils import AnalysisUtils
from .normalize_date_utils import NormalizeDateUtils
from .normalize_null_utils import NormalizeNullUtils
from .normalize_int_utils import NormalizeIntUtils
from .normalize_float_utils import NormalizeFloatUtils
from .normalize_string_utils import NormalizeStringUtils
from typing import Optional
import pandas as pd


class NormalizeDfUtils:

    @staticmethod
    def normalize_df(
        current_df: pd.DataFrame,
        null_values: list[any] = [],
        remove_accents: bool = True,
        keep_alphanum_space: bool = False,
        exclude_zero: bool = False,
        exclude_negative: bool = False
    ) -> None:
        columns: list[str] = current_df.columns.tolist()
        print(f"column names: {columns}")

        # Normaliza todas as strigs do current_df
        NormalizeStringUtils.normalize_string_by_columns(
            current_df=current_df,
            columns=columns,
            remove_accents=remove_accents,
            keep_alphanum_space=keep_alphanum_space
        )
        print("normalized: string")

        # Normaliza todas as representaçoes de valores ausentes convertendo em nan
        NormalizeNullUtils.normalize_null_by_columns(
            current_df=current_df,
            columns=columns,
            null_values=null_values
        )
        print("normalized: null")

        current_df_types: pd.DataFrame = AnalysisUtils.infer_column_types(current_df)
        print(f"column types: {current_df_types}")

        columns_int: list[str] = (
            current_df_types[current_df_types['type'] == 'int']['column_name'].tolist()
        )

        columns_float: list[str] = (
            current_df_types[current_df_types['type'] == 'float']['column_name'].tolist()
        )

        columns_time: list[str] = (
            current_df_types[current_df_types['type'] == 'time']['column_name'].tolist()
        )

        columns_date: list[str] = (
            current_df_types[current_df_types['type'] == 'date']['column_name'].tolist()
        )

        columns_bool: list[str] = (
            current_df_types[current_df_types['type'] == 'bool']['column_name'].tolist()
        )

        columns_datetime: list[str] = (
            current_df_types[current_df_types['type'] == 'datetime']['column_name'].tolist()
        )

        columns_none: list[str] = (
            current_df_types[current_df_types['type'] == 'None']['column_name'].tolist()
        )
        print(columns_none)

        NormalizeIntUtils.normalize_int_by_columns(
            current_df=current_df,
            columns=columns_int,
            exclude_zero=exclude_zero,
            exclude_negative=exclude_zero
        )
        print("normalized: int")

        NormalizeFloatUtils.normalize_float_by_columns(
            current_df=current_df,
            columns=columns_float,
            exclude_zero=exclude_zero,
            exclude_negative=exclude_zero
        )
        print("normalized: float")

        NormalizeDateUtils.normalize_str_date_to_date_standard_str_by_columns(
            current_df=current_df,
            columns=columns_date
        )
        print("normalized: date")

        NormalizeDateUtils.normalize_str_time_to_hh_mm_ss_by_columns(
            current_df=current_df,
            columns=columns_time
        )
        print("normalized: time")

        for null_column in columns_none:
            current_df.drop(null_column, axis=1, inplace=True)
        print("deleted columns: NoneType")

        # for nome_coluna in columns_str:
        #     df[nome_coluna] = df[nome_coluna].astype('string')

        print("normalized: finish")




import pandas as pd
import re
import unicodedata


class NormalizeStringUtils:
    # Função interna que normaliza uma string
    @staticmethod
    def _normalize_string(
        text: str,
        remove_accents: bool = False,
        keep_alphanum_space: bool = False
    ) -> str:
        '''
        Normaliza string:
        - Remove acentos (opcional)
        - Remove caracteres que não são letras, números ou espaço (opcional)
        - Remove espaços extras
        - Converte para UPPERCASE
        '''
        if not isinstance(text, str):
            return text

        if remove_accents:
            text = unicodedata.normalize('NFKD', text)
            text = ''.join(c for c in text if not unicodedata.combining(c))

        if keep_alphanum_space:
            text = re.sub(r'[^A-Za-z0-9 ]+', '', text)

        text = re.sub(r'\s+', ' ', text).strip()

        return text.upper()


    @staticmethod
    def normalize_string_by_columns(
        current_df: pd.DataFrame,
        columns: list[str],
        remove_accents: bool = False,
        keep_alphanum_space: bool = False
    ) -> None:
        '''
        Aplica a normalização de string nas colunas especificadas.
        '''
        for column in columns:
            if column not in current_df.columns:
                continue
            current_df[column] = current_df[column].apply(
                lambda row: NormalizeStringUtils._normalize_string(
                    row,
                    remove_accents=remove_accents,
                    keep_alphanum_space=keep_alphanum_space
                )
            )


    @staticmethod
    def normalize_unique_values_in_column(
        current_df: pd.DataFrame, 
        column: str, 
        target_value: str, 
        values_to_replace: list[str]
    ) -> None:
        '''
        Substitui valores específicos de uma coluna por um valor alvo.
        
        Parâmetros:
        - current_df: DataFrame contendo os dados.
        - column: Nome da coluna onde a substituição ocorrerá.
        - target_value: O valor que substituirá os valores encontrados.
        - values_to_replace: Lista de valores que serão substituídos pelo valor alvo.
        '''
        # Substitui os valores encontrados pela coluna `column` que estão na lista `values_to_replace` com `target_value`
        current_df[column].replace(values_to_replace, target_value, inplace=True)



from csv import Dialect
from datetime import datetime, time
from pandas.api.types import CategoricalDtype
from typing import Literal, Optional
from .regex_patterns import (
    DATE_REGEX_PATTERNS,
    TIME_REGEX_PATTERNS,
    DATE_TIME_REGEX_PATTERNS,
)
import csv
import numpy as np
import pandas as pd
import re


class AnalysisUtils:
    data_types = Literal['int', 'float', 'bool', 'date', 'time', 'datetime', 'str']

    @staticmethod
    def _number_cleaned(value: object) -> str:

        value_str: str = str(value).strip()

        if '.' in value_str and ',' in value_str:
            # formato 1000,00 -> 1000.00
            value_str: str = value_str.replace('.', '').replace(',', '.')

        if ',' in value_str:
            # formato 1.000,00 -> 1000.00
            value_str: str = value_str.replace(',', '.')

        return value_str


    @staticmethod
    def _is_zero(value: object) -> bool:
        cleaned_number_str: str = AnalysisUtils._number_cleaned(value)

        try:
            float_value: float = float(cleaned_number_str)

            if float_value == 0.0:
                return True
            else:
                return False
        except ValueError:
            return False


    @staticmethod
    def _is_int(value: object) -> bool:
        cleaned_number_str: str = AnalysisUtils._number_cleaned(value)

        try:
            float_value: float = float(cleaned_number_str)

            if float_value.is_integer():
                return True
            else:
                return False
        except ValueError:
            return False


    @staticmethod
    def _is_float(value: object) -> bool:
        '''
        Retorna True se o valor puder ser interpretado como float,
        incluindo formatos com separadores de milhar e vírgula como separador decimal.
        '''
        cleaned_number_str: str = AnalysisUtils._number_cleaned(value)

        try:
            float_value: float = float(cleaned_number_str)

            if not float_value.is_integer():
                return True
            else:
                return False
        except ValueError:
            return False


    @staticmethod
    def _is_bool(value: object) -> bool:
        """
        Retorna True se o valor puder ser interpretado como booleano.
        Considera 'true', 'false', '1', '0' (case-insensitive), além de bool nativo.
        """
        value_str = str(value).strip().lower()
        return value_str in {'True', 'False', 'true', 'false', }


    @staticmethod
    def _is_date(value: object) -> bool:
        value_str: str = str(value)

        for pattern, _ in DATE_REGEX_PATTERNS:
            if re.fullmatch(pattern, value_str):
                return True

        return False


    def _is_time(value: object) -> bool:
        value_str = str(value)

        for pattern, _ in TIME_REGEX_PATTERNS:
            if re.fullmatch(pattern, value_str):
                return True

        return False


    @staticmethod
    def _is_datetime(value: object) -> bool:
        """
        Verifica se o valor corresponde a um dos formatos datetime definidos em DATE_TIME_REGEX_PATTERNS.
        """

        value_str = str(value)

        for pattern, date_format in DATE_TIME_REGEX_PATTERNS:
            if re.fullmatch(pattern, value_str):
                try:
                    datetime.strptime(value_str, date_format)
                    return True
                except ValueError:
                    continue

        return False


    @staticmethod
    def _detect_type(value: object) -> Optional[Literal['int', 'float', 'bool', 'date', 'time', 'str', 'zero']]:
        if pd.isna(value):
            return None
        
        elif AnalysisUtils._is_zero(value):
            return 'zero'

        elif AnalysisUtils._is_int(value):
            return 'int'

        elif AnalysisUtils._is_float(value):
            return 'float'
        
        elif AnalysisUtils._is_time(value):
            return 'time'
        
        elif AnalysisUtils._is_date(value):
            return 'date'

        elif AnalysisUtils._is_bool(value):
            return 'bool'
        
        elif AnalysisUtils._is_datetime(value):
            return 'datetime'

        else:
            return 'str'


    @staticmethod
    def get_dominant_type(
        non_null_rows_sample: pd.Series,
        without_zero: bool = False
    ) -> Literal['int', 'float', 'bool', 'date', 'time', 'datetime', 'str', 'None', 'zero']:

        type_frequency: pd.Series = (
            non_null_rows_sample
            .apply(AnalysisUtils._detect_type)
            .value_counts()
        )

        if without_zero and 'zero' in type_frequency:
            type_frequency: pd.Series = type_frequency.drop(labels='zero')

        # Verifica a frequência de "float"
        float_count = type_frequency.get('float', 0)
        total_count = len(non_null_rows_sample)
        
        # TODO: Rever isso para analisar toda a coluna em caso de float e int
        # Se pelo menos 1% dos valores forem float, considera o tipo dominante como float
        if float_count / total_count >= 0.1:
            dominant_type = 'float'
        else:
            # Caso contrário, pega o tipo mais frequente
            dominant_type = type_frequency.idxmax()

        return dominant_type


    @staticmethod
    def sample_size(
        current_df: pd.DataFrame,
        rows_percent: int = 2
    ) -> int:
        if not (0 <= rows_percent <= 100):
            raise ValueError('rows_percent deve estar entre 0 e 100')

        row_count: int = len(current_df)
        sample_size: int = max(1, int((rows_percent / 100) * row_count))

        return sample_size


    @staticmethod
    def infer_column_types(
        current_df: pd.DataFrame,
        rows_percent: int = 2,
    ) -> pd.DataFrame:

        sample_size: int = AnalysisUtils.sample_size(current_df, rows_percent)

        columns: list[str] = current_df.columns.tolist()

        inferred_types: list[str] = []

        for column in columns:

            non_null_rows_sample: pd.Series = (
                current_df[column]
                .dropna()
                .head(sample_size)
            )

            if non_null_rows_sample.empty:
                dominant_type: str = 'None'
            else:
                dominant_type: str = AnalysisUtils.get_dominant_type(non_null_rows_sample)
                
                # Reanalisa as colunas numéricas caso haja dúvida
                if dominant_type == 'zero':
                    full_non_null_column: pd.Series = current_df[column].dropna()
                    dominant_type = AnalysisUtils.get_dominant_type(
                        non_null_rows_sample=full_non_null_column,
                        without_zero=True
                    )

            inferred_types.append(dominant_type)

        df_types: pd.DataFrame = pd.DataFrame({
            'column_name': columns,
            'type': inferred_types
        })
        print(df_types)

        return df_types


    @staticmethod
    def get_sep_delimiter(
        file_path: str,
        encoding: Optional[str] = 'latin1',
        skiprows: Optional[int] = 0
    ) -> str:
        sample_characters: int = 2048

        with open(file_path, 'r', encoding=encoding) as file:
            for _ in range(skiprows):
                next(file)  # pula as linhas iniciais

            sample: str = file.read(sample_characters)
            sniffer: csv.Sniffer = csv.Sniffer()
            dialect: Dialect = sniffer.sniff(sample)

        # with open(file_path, 'r', encoding=encoding) as file:
        #     sample: str = file.read(sample_characters)
        #     sniffer: csv.Sniffer = csv.Sniffer()
        #     dialect: Dialect = sniffer.sniff(sample)
        
        print(f'sep_delimiter: "{dialect.delimiter}"')
        return dialect.delimiter


    @staticmethod
    def get_frequency_by_column(current_df: pd.DataFrame, column: str) -> pd.DataFrame:
        '''
        Gera um current_df com a frequencia de valores de uma coluna
        '''
        current_df_frequency: pd.DataFrame = (
            current_df[column]
            .value_counts()
            .reset_index()
        )
    
        current_df_frequency.columns = [column, 'frequency']
    
        current_df_frequency: pd.DataFrame = (
            current_df_frequency
            .sort_values(
                by='frequency',
                ascending=False
            )
        )
    
        return current_df_frequency


    @staticmethod
    def get_unique_values(current_df: pd.DataFrame, column: str) -> pd.DataFrame:
        '''
        Retorna um dataframe com os valores únicos de uma coluna
        '''
        df_unique = pd.DataFrame({
            column: current_df[column].dropna().unique()
        })

        df_unique = df_unique.sort_values(
            by=column,
            ascending=True
        ).reset_index(drop=True)

        return df_unique
    
    @staticmethod
    def df_size_info(current_df: pd.DataFrame) -> None:
        return {
            "rows": current_df.shape[0],
            "columns": current_df.shape[1],
            "RAM MB": round(current_df.memory_usage(index=True, deep=True).sum() / (1024 * 1024), 2)
        }


from typing import Optional
import numpy as np
import pandas as pd


class NormalizeIntUtils:
    @staticmethod
    def _number_cleaned(value: object) -> str:

        value_str: str = str(value).strip()

        if '.' in value_str and ',' in value_str:
            # formato 1000,00 -> 1000.00
            value_str: str = value_str.replace('.', '').replace(',', '.')

        if ',' in value_str:
            # formato 1.000,00 -> 1000.00
            value_str: str = value_str.replace(',', '.')

        return value_str

    @staticmethod
    def _normalize_int(
        value: object,
        exclude_zero: bool = False,
        exclude_negative: bool = False
    ) -> Optional[int]:
        cleaned_number_str: str = NormalizeIntUtils._number_cleaned(value)

        try:
            value: int = int(float(cleaned_number_str))

            if (exclude_zero and value == 0) or (exclude_negative and value < 0):
                return pd.NA

            return value
        except (ValueError, TypeError):
            return pd.NA


    @staticmethod
    def normalize_int_by_columns(
        current_df: pd.DataFrame,
        columns: list[str],
        exclude_zero: bool = False,
        exclude_negative: bool = False
    ) -> None:
           
        for column in columns:
            if column in current_df.columns:
                current_df[column] = (
                    current_df[column]
                    .apply(
                        lambda row: NormalizeIntUtils._normalize_int(row, exclude_zero, exclude_negative)
                    )
                )



from .analysis_utils import AnalysisUtils
from .normalize_date_utils import NormalizeDateUtils
from .normalize_null_utils import NormalizeNullUtils
from .normalize_int_utils import NormalizeIntUtils
from .normalize_float_utils import NormalizeFloatUtils
from .normalize_string_utils import NormalizeStringUtils
from .normalize_df_utils import NormalizeDfUtils
from .file_analysis_utils import FileAnalysisUtils
from .regex_patterns import (
    RegexPatternsCombine,
    RAW_INVALID_VALUES,
    DATE_REGEX_PATTERNS,
    TIME_REGEX_PATTERNS,
    DATE_TIME_REGEX_PATTERNS,
    RAW_INVALID_VALUES
)


class Jiboia:

    # DATAFRAME
    normalize_df = staticmethod(NormalizeDfUtils.normalize_df)

    # TYPES:
    infer_column_types = staticmethod(AnalysisUtils.infer_column_types)

    # STRING
    normalize_string_by_columns = staticmethod(NormalizeStringUtils.normalize_string_by_columns)
    # normalize_category = staticmethod(NormalizeStringUtils.normalize_category)
    # normalize_unique_values = staticmethod(NormalizeStringUtils.normalize_unique_values)
    
    # NULL
    normalize_null_by_columns = staticmethod(NormalizeNullUtils.normalize_null_by_columns)

    # BOOL
    # normalize_bool = staticmethod(GPUNormalizeBooleanUtils.normalize_bool)
    
    # NUMBER
    normalize_int_by_columns = staticmethod(NormalizeIntUtils.normalize_int_by_columns)
    normalize_float_by_columns = staticmethod(NormalizeFloatUtils.normalize_float_by_columns)


    # DATE
    normalize_str_date_to_date_standard_str_by_columns = staticmethod(NormalizeDateUtils.normalize_str_date_to_date_standard_str_by_columns)
    # normalize_datetime = staticmethod(NormalizeDateUtils.normalize_datetime)
    # is_date_column_by_samples = staticmethod(NormalizeDateUtils.is_date_column_by_samples)
    # is_date = staticmethod(NormalizeDateUtils.is_date)
    # to_date_iso = staticmethod(NormalizeDateUtils.to_date_iso)
    # to_datetime = staticmethod(NormalizeDateUtils.to_datetime)


    # TIME
    normalize_str_time_to_hh_mm_ss_by_columns = staticmethod(NormalizeDateUtils.normalize_str_time_to_hh_mm_ss_by_columns)
    
    # normalize_time = staticmethod(GPUNormalizeTimeUtils.normalize_time)
    # is_time = staticmethod(GPUNormalizeTimeUtils.is_time)

    # # TIMEDELTA
    # normalize_timedelta = staticmethod(GPUNormalizeTimeUtils.normalize_timedelta)

    # # DATETIME
    # create_datetime_column_from_date_and_time = staticmethod(GPUNormalizeDateTimeUtils.create_datetime_column_from_date_and_time)
    # create_day_week_column = staticmethod(GPUNormalizeDateTimeUtils.create_day_week_column)
    # create_month_br_category_column = staticmethod(GPUNormalizeDateTimeUtils.create_month_br_category_column)
    # create_day_week_br_category_column = staticmethod(GPUNormalizeDateTimeUtils.create_day_week_br_category_column)
    # create_hour_category_column = staticmethod(GPUNormalizeDateTimeUtils.create_hour_category_column)

    # ANALYSIS
    # drop_columns = staticmethod(AnalysisUtils.drop_columns)
    # cudf_size_info = staticmethod(AnalysisUtils.cudf_size_info)
    # pandasdf_size_info = staticmethod(AnalysisUtils.pandasdf_size_info)
    # get_gpu_memory_info = staticmethod(AnalysisUtils.get_gpu_memory_info)
    # is_vram_use_limit = staticmethod(AnalysisUtils.is_vram_use_limit)
    # get_index_samples = staticmethod(AnalysisUtils.get_index_samples)
    # infer_by_sample = staticmethod(AnalysisUtils.infer_by_sample)

    # # FILE
    get_csv_sep_delimiter = staticmethod(FileAnalysisUtils.get_csv_sep_delimiter)
    # get_csv_info = staticmethod(FileAnalysisUtils.get_csv_info)
    # convert_csv_to_utf8 = staticmethod(FileAnalysisUtils.convert_csv_to_utf8)
    # convert_all_csvs_to_utf8 = staticmethod(FileAnalysisUtils.convert_all_csvs_to_utf8)
    # read_csv_files = staticmethod(FileAnalysisUtils.read_csv_files)

# instância única para ser usada como namespace
jiboia = Jiboia()

__all__ = ["jiboia", "Jiboia"]



from csv import Dialect
from datetime import datetime, time
from pandas.api.types import CategoricalDtype
from typing import Literal, Optional
from .regex_patterns import (
    DATE_REGEX_PATTERNS,
    TIME_REGEX_PATTERNS,
    DATE_TIME_REGEX_PATTERNS,
)
import csv
import numpy as np
import pandas as pd
import re


import chardet


class FileAnalysisUtils:
    @staticmethod
    def get_csv_sep_delimiter(
        file_path: str,
        encoding: Optional[str] = 'latin1',
        skiprows: Optional[int] = 0
    ) -> str:
        sample_characters: int = 2048

        with open(file_path, 'r', encoding=encoding) as file:
            for _ in range(skiprows):
                next(file)  # pula as linhas iniciais

            sample: str = file.read(sample_characters)
            sniffer: csv.Sniffer = csv.Sniffer()
            dialect: Dialect = sniffer.sniff(sample)

        with open(file_path, "rb") as f:
            rawdata = f.read(100000)  # lê parte do arquivo
            result = chardet.detect(rawdata)
            print(result)

        print(f'sep_delimiter: "{dialect.delimiter}"')
        return dialect.delimiter



